#include "db/records.h"
#include "jwt/jwt.h"
#include "server/response.h"
#include "server/study.h"
#include "server/types.h"
#include "server/util.h"


#include "db/auth.h"

#include "hash/hash.h"

#include <civetweb.h>

#include <cjson/cJSON.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>

typedef struct
{
    char* action;
    int id;
    int has_id;
    char* user_id;
    int problem_id;
    int has_problem_id;
} record_form_t;

static void record_form_dtor(record_form_t* form)
{
    if (form->action) free(form->action);
    if (form->user_id) free(form->user_id);
}

static int field_found(const char* key, const char* filename, char* path, size_t pathlen, void* user_data)
{
    return MG_FORM_FIELD_HANDLE_GET;
}

static int field_get(const char* key, const char* value, size_t valuelen, void* user_data)
{
    record_form_t* form = (record_form_t*)user_data;
    if (strcmp(key, "action") == 0) {
        form->action = kqm_strndup(value, valuelen);
    } else if (strcmp(key, "id") == 0) {
        char* id_str = kqm_strndup(value, valuelen);

        form->id     = atoi(id_str);
        form->has_id = 1;
        free(id_str);
    } else if (strcmp(key, "user_id") == 0) {
        form->user_id = kqm_strndup(value, valuelen);
    } else if (strcmp(key, "problem_id") == 0) {
        char* id_str = kqm_strndup(value, valuelen);

        form->problem_id     = atoi(id_str);
        form->has_problem_id = 1;
        free(id_str);
    }
    return MG_FORM_FIELD_HANDLE_GET;
}

static void impl_query(mg_connection* conn, record_form_t* form)
{
    if (!form->has_id) {
        res_need_xxx(conn, "record id");
        return;
    }

    char* result = NULL;
    int flag     = get_record(form->id, &result);
    if (!flag) {
        res_500(conn, "failed to get a record");
        return;
    }
    if (!result) {
        res_404(conn, "the record does not exist");
        return;
    }

    res_get_record(conn, result);
    free(result);
}

static void impl_all(mg_connection* conn, record_form_t* form)
{
    char* result = NULL;
    int flag     = all_records(&result);
    if (!flag) {
        res_500(conn, "failed to get all records");
        return;
    }

    res_all_records(conn, result);
    free(result);
}

static void impl_by(mg_connection* conn, record_form_t* form)
{
    if (!form->user_id) {
        res_need_xxx(conn, "user id");
        return;
    }

    char* result = NULL;
    int flag     = all_records_by(form->user_id, &result);
    if (!flag) {
        res_500(conn, "failed to get the records by the user");
        return;
    }

    res_all_records_by(conn, result);
    free(result);
}

static void impl_of(mg_connection* conn, record_form_t* form)
{
    if (!form->has_problem_id) {
        res_need_xxx(conn, "problem id");
        return;
    }

    char* result = NULL;
    int flag     = all_records_of(form->problem_id, &result);
    if (!flag) {
        res_500(conn, "failed to get the records of the problem");
        return;
    }

    res_all_records_of(conn, result);
    free(result);
}

static void impl_count(mg_connection* conn)
{
    int count = records_count();
    res_records_count(conn, count);
}

static void impl_have(mg_connection* conn, record_form_t* form)
{
    if (!form->user_id) {
        res_need_xxx(conn, "user id");
        return;
    }
    if (!form->has_problem_id) {
        res_need_xxx(conn, "problem id");
        return;
    }

    int status = have_ac(form->user_id, form->problem_id);
    res_have_ac(conn, status);
}

int records_handler(mg_connection* conn, void* cbdata)
{
    const mg_request_info* post_body = mg_get_request_info(conn);

    if (post_body == NULL) {
        res_null_req(conn);
        return 1;
    }

    if (strcmp(post_body->request_method, "POST")) {
        res_must_post(conn);
        return 1;
    }

    record_form_t form = {NULL, 0, 0, NULL, 0, 0};

    mg_form_data_handler record_callback = {
        .field_found = field_found,
        .field_get   = field_get,
        .field_store = NULL,
        .user_data   = &form,
    };

    mg_handle_form_request(conn, &record_callback);


    if (!form.action) {
        res_need_action(conn);
    } else {
        if (!strcmp(form.action, "query")) {
            impl_query(conn, &form);
        } else if (!strcmp(form.action, "by")) {
            impl_by(conn, &form);
        } else if (!strcmp(form.action, "of")) {
            impl_of(conn, &form);
        } else if (!strcmp(form.action, "count")) {
            impl_count(conn);
        } else if (!strcmp(form.action, "have")) {
            impl_have(conn, &form);
        } else {
            res_bad_action(conn);
        }
    }

    record_form_dtor(&form);
    return 1;
}
